/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;

import org.ow2.easywsdl.schema.util.XMLPrettyPrinter;
import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.InterfaceType;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.Util.MappingUuidThreesome;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.ebmwebsourcing.wsstar.addressing.definition.WSAddressingFactory;
import com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType;
import com.ebmwebsourcing.wsstar.addressing.definition.api.ReferenceParametersType;
import com.ebmwebsourcing.wsstar.addressing.definition.api.WSAddressingException;
import com.ebmwebsourcing.wsstar.notification.definition.WSNotificationFactory;
import com.ebmwebsourcing.wsstar.notification.definition.utils.WSNotificationException;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.WsnbConstants;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.CreatePullPoint;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.CreatePullPointResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.FilterType;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.GetCurrentMessage;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.GetCurrentMessageResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Message;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.NotificationMessageHolderType;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Notify;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Renew;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.RenewResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Subscribe;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.SubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.TopicExpressionType;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Unsubscribe;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.UnsubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.DestroyRegistration;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.DestroyRegistrationResponse;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.RegisterPublisher;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.RegisterPublisherResponse;
import com.ebmwebsourcing.wsstar.notification.definition.inout.WSNotificationReader;
import com.ebmwebsourcing.wsstar.notification.definition.inout.WSNotificationWriter;
import com.ebmwebsourcing.wsstar.notification.extension.api.ResourcesUuidType;
import com.ebmwebsourcing.wsstar.notification.extension.api.SOAParameterType;
import com.ebmwebsourcing.wsstar.notification.extension.utils.WSNotificationExtensionException;
import com.ebmwebsourcing.wsstar.notification.extension.WSNotificationExtensionFactory;
import com.ebmwebsourcing.wsstar.notification.extension.utils.WsnSpecificTypeHelper;
import com.ebmwebsourcing.wsstar.notification.service.basenotification.impl.CreatePullPointMgr;
import com.ebmwebsourcing.wsstar.notification.service.basenotification.impl.NotificationProducerMgr;
import com.ebmwebsourcing.wsstar.notification.service.basenotification.impl.PullPointMgr;
import com.ebmwebsourcing.wsstar.notification.service.basenotification.impl.SubscriptionManagerMgr;
import com.ebmwebsourcing.wsstar.notification.service.brokerednotification.WsnbrNotificationBroker;
import com.ebmwebsourcing.wsstar.notification.service.brokerednotification.WsnbrPublisherRegistrationManager;
import com.ebmwebsourcing.wsstar.notification.service.brokerednotification.impl.RegistrationsMgr;
import com.ebmwebsourcing.wsstar.notification.service.fault.WSNotificationFault;
import com.ebmwebsourcing.wsstar.notification.service.persistence.WsnPersistence;
import com.ebmwebsourcing.wsstar.notification.service.topic.WstopTopicManager;

/**
 * The WS-BrokeredNotification service implementation
 * 
 * @author tdejean - eBMWebsourcing
 * 
 */
public class BrokeredNotificationService implements WsnbrNotificationBroker,
WsnbrPublisherRegistrationManager {


    private final static String persistedRegistrationFolderName = "SavePublisherRegistrationRPs" + File.separatorChar;
    private final static String persistedSubscriptionFolderName = "SaveSubscriptionManagerRPs" + File.separatorChar;
    private final static String forwardedSubscribeResponsesFolderName = "SaveProducerSubscribeResponses" + File.separatorChar;
    
    private final static String serializedThreesomesFolderName = "SaveSubscriptionMappingThreesomes" + File.separatorChar;
       
    private File forwardedSubscribeResponsesFolder = null;
    private File subscriptionMappingFolder = null;
    
    private final String NOTIFICATION_BROKER_SERVICE_NAME = "NotificationBrokerService";

    private final String PUBLISHER_REGISTRATION_MANAGER_SERVICE_NAME = "PublisherRegistrationManagerService";

    private final String SUBSCRIPTION_MANAGER_SERVICE_NAME = "SubscriptionManagerService";

    private final Logger logger;

    private NotificationProducerMgr wsnbProducerService = null;

    private SubscriptionManagerMgr wsnbSubscriptionManagerService = null;

    private RegistrationsMgr wsnbrRegistrationMgr = null;

    private CreatePullPointMgr wsnbCreatePullPointService = null;

    private PullPointMgr wsnbPullPointService = null;

    private WstopTopicManager wstopTopicsMgr = null;

    private DeliveryChannel dc = null;

    private ComponentContext cptContext = null;

    //private final Map<String, List<RemoteSubscriptionProducerRefs>> forwardedSubscriptionRefs;
    
    /**
     * key : consumer Subscription uuids
     * value : pair <"publisher(producer) Registration uuid", "producer SubscribeResponse uuid">
     */
    private List<MappingUuidThreesome> subscriptionMappingList = null;    
    
    private String[] additonalnsAndPrefixMapping = null;

    private WsnPersistence wsnPersistenceMgr = null;
    
    // ---- save subscribe Response received from producer
    private Map<String,SubscribeResponse> receivedSubsResponses = null;
    
    /**
     * Constructor ...
     * 
     * @param logger
     * @param component
     * @param additonalnsAndPrefixMapping
     */
    public BrokeredNotificationService(final Logger logger, final AbstractComponent component,
            final String[] additonalnsAndPrefixMapping) {
        super();
        this.logger = logger;
        this.dc = component.getChannel();
        this.cptContext = component.getContext();
        String componentWorkingDir = this.cptContext.getWorkspaceRoot();
       
        this.additonalnsAndPrefixMapping = additonalnsAndPrefixMapping;

        //this.forwardedSubscriptionRefs = Collections.synchronizedMap(new HashMap<String, List<RemoteSubscriptionProducerRefs>>());
        this.subscriptionMappingList = Collections.synchronizedList(new ArrayList<MappingUuidThreesome>());        
        this.receivedSubsResponses = Collections.synchronizedMap(new HashMap<String,SubscribeResponse>());

        try {

            this.wstopTopicsMgr = new WstopTopicManager(BrokeredNotificationService.class
                    .getResourceAsStream("/SupportedTopicsSet.xml"));

            this.forwardedSubscribeResponsesFolder = new File(componentWorkingDir+File.separatorChar+forwardedSubscribeResponsesFolderName);
            
            this.subscriptionMappingFolder = new File(componentWorkingDir+File.separatorChar+serializedThreesomesFolderName);
            
            this.wsnPersistenceMgr = new WsnPersistence(new File(componentWorkingDir+File.separatorChar+persistedSubscriptionFolderName),
                    new File(componentWorkingDir+File.separatorChar+persistedRegistrationFolderName));
            
            this.wsnbSubscriptionManagerService = new SubscriptionManagerMgr(logger,
                    this.wstopTopicsMgr,this.wsnPersistenceMgr);

            final QName notifSubscriptionManagerServiceQName = new QName(component
                    .getNativeWsdl()
                    .getDescription().getTargetNamespace(),
                    this.SUBSCRIPTION_MANAGER_SERVICE_NAME);
            final InterfaceType subsManagerIntf = component.getNativeWsdl().getDescription()
            .getService(
                    notifSubscriptionManagerServiceQName).getInterface();
            final Endpoint subsManagerEdpt = component.getNativeWsdl().getDescription()
            .findEndpointsImplementingInterface(subsManagerIntf).get(0);
            this.wsnbSubscriptionManagerService.setSubscriptionsManagerEdp(subsManagerEdpt
                    .getName());
            this.wsnbSubscriptionManagerService
            .setSubscriptionsManagerService(notifSubscriptionManagerServiceQName);
            this.wsnbSubscriptionManagerService.setSubscriptionsManagerInterface(subsManagerIntf
                    .getQName());

            this.wsnbProducerService = new NotificationProducerMgr(logger, this.wstopTopicsMgr,
                    this.wsnbSubscriptionManagerService);

            final QName notifBrokerServiceQName = new QName(component.getNativeWsdl()
                    .getDescription()
                    .getTargetNamespace(), this.NOTIFICATION_BROKER_SERVICE_NAME);
            final InterfaceType notifBrokerIntf = component.getNativeWsdl().getDescription()
            .getService(
                    notifBrokerServiceQName).getInterface();
            final Endpoint notifBrokerEdpt = component.getNativeWsdl().getDescription()
            .findEndpointsImplementingInterface(notifBrokerIntf).get(0);
            this.wsnbProducerService.setNotificationProducerEdp(notifBrokerEdpt.getName());
            this.wsnbProducerService.setNotificationProducerService(notifBrokerServiceQName);
            this.wsnbProducerService.setNotificationProducerInterface(notifBrokerIntf.getQName());

            this.wsnbrRegistrationMgr = new RegistrationsMgr(logger, this.wstopTopicsMgr,this.wsnPersistenceMgr);
            final QName pubRegManServiceQName = new QName(component.getNativeWsdl()
                    .getDescription()
                    .getTargetNamespace(), this.PUBLISHER_REGISTRATION_MANAGER_SERVICE_NAME);
            final InterfaceType pubRegManIntf = component.getNativeWsdl().getDescription()
            .getService(
                    pubRegManServiceQName).getInterface();
            final Endpoint pubRegManEdpt = component.getNativeWsdl().getDescription()
            .findEndpointsImplementingInterface(pubRegManIntf).get(0);
            this.wsnbrRegistrationMgr
            .setPublisherRegistrationMgrServiceQName(pubRegManServiceQName);
            this.wsnbrRegistrationMgr
            .setPublisherRegistrationMgrEdpAddress(pubRegManEdpt.getName());
            this.wsnbrRegistrationMgr.setPublisherRegistrationMgrInterfaceQName(pubRegManIntf
                    .getQName());

            this.wsnbrRegistrationMgr.setNotificationBrokerEdp(notifBrokerEdpt.getName());
            this.wsnbrRegistrationMgr.setNotificationBrokerService(notifBrokerServiceQName);
            this.wsnbrRegistrationMgr.setNotificationBrokerInterface(notifBrokerServiceQName);

            this.wsnbPullPointService = new PullPointMgr(logger);
            this.wsnbCreatePullPointService = new CreatePullPointMgr(logger, this.wsnbPullPointService);

        } catch (final WSDLException e) {
            e.printStackTrace();
        }
    }
 
    /**
     * WstopTopicMgr getter
     * @return
     */
    public WstopTopicManager getWstopTopicsMgr() {
        return this.wstopTopicsMgr;
    }
  
    /**
     * restore previous context (subscriptions, registrations,
     * mapping between "consumer subscrption"/"producers subscription response", ...) 
     *      
     * @throws WSNotificationException
     * @throws WSNotificationFault
     */
    public void restorePreviousState() throws WSNotificationException, WSNotificationFault{

        this.wsnbSubscriptionManagerService.restorePersistedSubscriptions();
        this.wsnbrRegistrationMgr.restorePersistedRegistration();
        this.receivedSubsResponses = this.getSubscribeResponseToRestore();
        this.subscriptionMappingList = this.deserializeSubscriptionMappingThreesomes();
        
        if (this.receivedSubsResponses.size() == 0)
            this.reInitForwardSubscribeRequests();
    }
    
    /**
     * call to cleanly stop the se-notification 
     * @throws WSNotificationException 
     */
    public void unsubscribeToAllProducers() throws WSNotificationException {

        try {
            for (Map.Entry<String, SubscribeResponse> mapPairItem : this.receivedSubsResponses.entrySet()) {

                String currentUuid = mapPairItem.getKey();

                String curAddress =  mapPairItem.getValue().getSubscriptionReference().getAddress();
                ReferenceParametersType curRefParam = mapPairItem.getValue().getSubscriptionReference().getReferenceParameters();

                final SOAParameterType curSOAParam = WsnSpecificTypeHelper.getSOAParameter(curRefParam);

                final QName targetService = curSOAParam.getService();
                final QName targetInterface = curSOAParam.getInterface();
                final String targetEdp = curSOAParam.getEndpoint();           

                this.logger.log(Level.FINE, "target producer Service to unsubscribe to : "  + targetService + "\n");
                this.logger.log(Level.FINE, "target producer Interface to unsubscribe to : " + targetInterface + "\n");
                this.logger.log(Level.FINE, "target producer Endpoint to unsubscribe to : " + targetEdp + "\n");

                // Replace Notification Sender Endpoint Reference
                // with notification broker's one
                // and forword the existaing registrationId
                final Unsubscribe unsubscribeToForward = WSNotificationFactory.getInstance().createUnsubscribe();

                final ResourcesUuidType resourceUuids = WSNotificationExtensionFactory.getInstance().createResourcesUuidType();

                resourceUuids.addUuid(currentUuid);

                WsnSpecificTypeHelper.setResourcesUuidType(resourceUuids, unsubscribeToForward);

                final com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType edpRef = WSAddressingFactory
                .getInstance().newEndpointReferenceType();
                //edpRef.setAddress(regRefItem.getAddress());
                edpRef.setAddress(curAddress);

                WsnSpecificTypeHelper.setEndpointReference(edpRef, unsubscribeToForward);

                // ---- here, i forward received notifications to
                // subscribed consumer
                final Document requestAsDocument = WSNotificationWriter.getInstance(this.additonalnsAndPrefixMapping)
                .writeUnsubscribe(unsubscribeToForward/* cancellation */);
                // As the same registrationId is used the return is
                // not necessary (?)
                this.forwardUnsubscriptionRequest(targetService, targetInterface, targetEdp, requestAsDocument);
                
                // remove persisted and mapping producer subscribeResponse 
                this.removePersistedSubscribeResponse(currentUuid);                
                this.removeThreesomesByProducerSubscribeResponseUuid(currentUuid);                
            }
            // remove producer subscribe response from list
            this.receivedSubsResponses.clear();
        }  catch (WSNotificationExtensionException e) {
            throw new WSNotificationException(e);
        } catch (WSAddressingException e) {
            throw new WSNotificationException(e);
        }
    }
    
    /**
     * @throws WSNotificationFault 
     * @throws WSNotificationException 
     * 
     */
    public void reInitForwardSubscribeRequests() throws WSNotificationFault, WSNotificationException{
        
        Map<String, Subscribe> subscribeToforward = 
            this.wsnbSubscriptionManagerService.getSubscriptionsRPAsSubscribePayoad();

        for (Map.Entry<String, Subscribe> mapPairItem : subscribeToforward.entrySet()) {
            this.subscribe(mapPairItem.getValue(), true,mapPairItem.getKey());                
        }

        
    }
    // ##################### Forwarding requests methods #######################
    
    /**
     * forward producer notify requests to consumer 
     * 
     * @param subscribedConsumerService
     * @param subscribedConsumerInterface
     * @param subscribedConsumerEdp
     * @param request
     */           
    private void forwardNotification(final QName subscribedConsumerService,
            final QName subscribedConsumerInterface, final String subscribedConsumerEdp, final Document request) {

        //final QName NOTIFICATION_CONSUMER_SERVICE_QNAME = new QName(
        //        "http://www.ebmwebsourcing.com/WS-BaseNotification", "NotificationConsumerService");

        InOnly inOnlyExchange;
        final ServiceEndpoint serviceEndpoint = this.cptContext.getEndpoint(subscribedConsumerService,
                subscribedConsumerEdp);

        try {
            inOnlyExchange = this.dc.createExchangeFactory(serviceEndpoint).createInOnlyExchange();
            final NormalizedMessage normalizedMessage = inOnlyExchange.createMessage();
            normalizedMessage.setContent(new DOMSource(request));
            inOnlyExchange.setInMessage(normalizedMessage);
            inOnlyExchange.setInterfaceName(subscribedConsumerInterface);
            inOnlyExchange.setService(subscribedConsumerService);
            inOnlyExchange.setOperation(WsnbConstants.NOTIFY_QNAME);

            this.dc.sendSync(inOnlyExchange);
            //closeExchange(inOnlyExchange);
        } catch (final MessagingException e) {
            this.logger
            .log(Level.WARNING,
            "An error occurs during messageExchange building and sending process, for \"notify\" request");
            // e.printStackTrace();
        }
    }

    /**
     * Used to forward a consumer Subscribe request from the Broker to the
     * Producer
     * 
     * @param subscribedProducerEdp
     * @param request
     * @return
     */
    private SubscribeResponse forwardSubscriptionRequest(final QName registredProducerService,
            final QName registredProducerInterface, final String registredProducerEdp, final Document request) {
        SubscribeResponse result = null;
        Document subsResp = null;

        //final QName NOTIFICATION_PRODUCER_SERVICE_QNAME = new QName(
        //        "http://www.ebmwebsourcing.com/WS-BaseNotification", "NotificationProducerService");
        InOut inOutExchange;

        final ServiceEndpoint serviceEndpoint = this.cptContext.getEndpoint(registredProducerService,
                registredProducerEdp);

        try {
            inOutExchange = this.dc.createExchangeFactory(serviceEndpoint).createInOutExchange();

            final NormalizedMessage normalizedMessage = inOutExchange.createMessage();

            normalizedMessage.setContent(new DOMSource(request)); 
            
            
            inOutExchange.setInMessage(normalizedMessage);
            inOutExchange.setInterfaceName(registredProducerInterface);
            inOutExchange.setService(registredProducerService);
            inOutExchange.setOperation(WsnbConstants.SUBSCRIBE_QNAME);
            
            if (this.dc.sendSync(inOutExchange)) {
                
                
                
                if (inOutExchange.getOutMessage() != null) {
                    subsResp = UtilFactory.getSourceUtil().createDocument(
                            inOutExchange.getOutMessage().getContent());
                    
                    
                    result = WSNotificationReader.getInstance().readSubscribeResponse(subsResp);                                        
                    
                } else if (inOutExchange.getFault() != null) {
                    subsResp = UtilFactory.getSourceUtil().createDocument(inOutExchange.getFault().getContent());
                    this.logger.log(Level.WARNING,
                    " [forwardSubscriptionRequest(...) ] - A WS-Notification Fault have been return on \"subscribe\" request !!");
                }

            }
            this.closeExchange(inOutExchange);
        } catch (final MessagingException e) {
            this.logger
            .log(
                    Level.WARNING,
                    "["
                    + e.getClass().getSimpleName()
                    + "] -An error occurs during messageExchange building and sending process, for \"subscribe\" request");
            // logger.log(Level.WARNING,"The \"Subscribe\" request can not be forwarded because the endpoint does not still exit !");
        } catch (final PEtALSCDKException e) {
            this.logger
            .log(
                    Level.WARNING,
                    "["
                    + e.getClass().getSimpleName()
                    + "] - An error occurs during messageExchange building and sending process, for \"subscribe\" request");
            // e.printStackTrace();
        } catch (final WSNotificationException e) {
            this.logger
            .log(
                    Level.WARNING,
                    "["
                    + e.getClass().getSimpleName()
                    + "] -An error occurs during messageExchange building and sending process, for \"subscribe\" request");
            // e.printStackTrace();
        }
        return result;
    }

    /**
     * Used to forward a consumer Unsubscription request from the Broker to the
     * Producer
     * 
     * @param subscribedProducerEdp
     * @param request
     * @return
     */
    // private Document forwardUnsubscriptionRequest(String
    // subscribedProducerEdp, Document request){
    private Document forwardUnsubscriptionRequest(final QName registredProducerService,
            final QName registredProducerInterface, final String registredProducerEdp, final Document request) {
        Document result = null;

        //final QName SUBSCRIPTION_MANAGER_SERVICE_QNAME = new QName(
        //        "http://www.ebmwebsourcing.com/WS-BaseNotification", "SubscriptionManagerService");

        InOut inOutExchange;
        // ServiceEndpoint serviceEndpoint =
        // this.cptContext.getEndpoint(SUBSCRIPTION_MANAGER_SERVICE_QNAME,
        // subscribedProducerEdp);
        final ServiceEndpoint serviceEndpoint = this.cptContext.getEndpoint(registredProducerService,
                registredProducerEdp);

        try {
            inOutExchange = this.dc.createExchangeFactory(serviceEndpoint).createInOutExchange();

            final NormalizedMessage normalizedMessage = inOutExchange.createMessage();

            normalizedMessage.setContent(new DOMSource(request));
            inOutExchange.setInMessage(normalizedMessage);
            inOutExchange.setInterfaceName(registredProducerInterface);
            inOutExchange.setService(registredProducerService);
            inOutExchange.setOperation(WsnbConstants.UNSUBSCRIBE_QNAME);

            if (this.dc.sendSync(inOutExchange)) {
                if (inOutExchange.getOutMessage() != null) {
                    result = UtilFactory.getSourceUtil().createDocument(
                            inOutExchange.getOutMessage().getContent());
                } else if (inOutExchange.getFault() != null) {
                    result = UtilFactory.getSourceUtil().createDocument(
                            inOutExchange.getFault().getContent());
                    this.logger
                    .log(
                            Level.WARNING,
                    " [forwardUnsubscriptionRequest(...) ] - A WS-Notification Fault have been return on \"subscribe\" request !!");
                }
            }
            this.closeExchange(inOutExchange);
        } catch (final MessagingException e) {
            // e.printStackTrace();
            // logger.log(Level.WARNING,"The \"unsubscribe\" request can not be forwarded because the endpoint does not still exit !");
            this.logger
            .log(
                    Level.WARNING,
                    "["
                    + e.getClass().getSimpleName()
                    + "] -An error occurs during messageExchange building and sending process, for \"unsubscribe\" request");
        } catch (final PEtALSCDKException e) {
            // e.printStackTrace();
            // logger.log(Level.WARNING,"The \"unsubscribe\" request can not be forwarded because the endpoint does not still exit !");
            this.logger
            .log(
                    Level.WARNING,
                    "["
                    + e.getClass().getSimpleName()
                    + "] -An error occurs during messageExchange building and sending process, for \"unsubscribe\" request");
        }
        return result;
    }
    
    /**
     * Close the exchange
     * 
     * @param exchange
     *            The exchange to close
     */
    private void closeExchange(final MessageExchange exchange) {
        // close the messageExchange with the JBI NMR
        try {
            if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
                exchange.setStatus(ExchangeStatus.DONE);
                this.dc.send(exchange);
            }
        } catch (final MessagingException e) {
            final String errorMsg = "Error while closing JBI MessageExchange.";
            this.logger.log(Level.WARNING, errorMsg, e);
        }
    }
    // #####################################################################################
    // ################# methods that implement WS-Notification operations #################
    
    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.engine.test.BrokeredNotification#notify(com.ebmwebsourcing
     * .wsstar.notification.base.Notify)
     */
    public void notify(final Notify request) {
        // forward notification to NotificationconsumerEngine :
        // Not used
        try {
            this.logger.log(Level.FINE, "I am the \"notification broker\" and i have received"
                    + " a notification from a registred notification producer !");
            this.logger.log(Level.FINE, "I must forward this notification to all consumer" +
            		" that have subscribed to it");

            // --- create a new Endpoint Reference to substitute to notification
            // msg 's one ---
            EndpointReferenceType newNotificationMsgOwner;
            newNotificationMsgOwner = WSAddressingFactory.getInstance().newEndpointReferenceType();
            final ReferenceParametersType refParam = newNotificationMsgOwner.newReferenceParameters();
            newNotificationMsgOwner.setReferenceParameters(refParam);

            final SOAParameterType soaParam = WSNotificationExtensionFactory.getInstance().createSOAParameterType();
            soaParam.setEndpoint(this.wsnbProducerService.getNotificationProducerEdp());
            soaParam.setService(this.wsnbProducerService.getNotificationProducerService());
            soaParam.setInterface(this.wsnbProducerService.getNotificationProducerInterfaceQName());
            WsnSpecificTypeHelper.setSOAParameter(soaParam, refParam);

            newNotificationMsgOwner.setAddress(SOAParameterType.DEFAULT_PETALS_ADDRESS);
            // ----------------------------------------------------------------------------------

            final List<NotificationMessageHolderType> messages = request.getNotificationMessage();

            for (final NotificationMessageHolderType msg : messages) {

                if (this.logger.isLoggable(Level.FINE) && msg != null) {
                    final Message mess = msg.getMessage();
                    if (mess != null){
                        final Element messContent = WsnSpecificTypeHelper.getContentFromMessage(mess);
                        if (messContent != null) {
                            this.logger.log(Level.FINE,"\nNotification message content : " + XMLPrettyPrinter.prettyPrint(messContent.getOwnerDocument()));
                        }
                    }
                }

                List<String> subscriptionIds = null;

                final EndpointReferenceType subsEdp = msg.getSubscriptionReference();

                // ---------------------------------------------------------------------------------------------------------
                // -------- create the list of subscriptions (subscription id) which notification must be fowarded to : ----
                // ---------------------------------------------------------------------------------------------------------
                if (subsEdp!=null) {
                    final String subsUuid = WsnSpecificTypeHelper.getResourcesUuidType(subsEdp.getReferenceParameters()).getUuids().get(0);
                    this.logger.log(Level.FINE, "subscription's Uuid associated to notify payload : " + subsUuid);
                    
                    subscriptionIds = new ArrayList<String>();                    
                    // --- get all subscriptionIds from subscriptionMappingList   
                    for (MappingUuidThreesome threesomeItem : this.subscriptionMappingList) {
                        subscriptionIds.add(threesomeItem.getConsumerSubcriptionUuid());
                    }
                                        
                    /*
                    List<RemoteSubscriptionProducerRefs> currentAssociatedRemoteSubs = null;

                    for (final String consSubsRefUuidsItem : this.forwardedSubscriptionRefs.keySet()) {
                        currentAssociatedRemoteSubs = this.forwardedSubscriptionRefs.get(consSubsRefUuidsItem);
                        for (final RemoteSubscriptionProducerRefs remoteSubscriptionProducerRefsItem : currentAssociatedRemoteSubs) {

                            final String currentUuidToCheck = remoteSubscriptionProducerRefsItem.getSubscriptionIdOfSubsRef();

                            if (currentUuidToCheck.compareTo(subsUuid) == 0 && ! subscriptionIds.contains(consSubsRefUuidsItem)) {
                                subscriptionIds.add(consSubsRefUuidsItem);
                            }
                        }
                    }*/


                    this.logger.log(Level.FINE, "consumers subscription's Uuid found : " + subscriptionIds);

                } else {
                    final TopicExpressionType topic = msg.getTopic();
                    for (final QName nsToDisplay : topic.getTopicNameSpace()) {
                        this.logger.log(Level.FINE, "Topic namespace : " + nsToDisplay.toString());
                    }
                    this.logger.log(Level.FINE, "Topic dialect : " + topic.getDialect());
                    this.logger.log(Level.FINE, "Topic content : " + topic.getContent());

                    // Get, from the "Topic manager", the subscriptionIds of the
                    // given topic
                    subscriptionIds = this.wstopTopicsMgr.getSubscriptionIdsFromTopicsSet(
                            topic, true);

                    // Get from the "Subscription manager" the associated consumers
                    // list
                    this.logger
                    .log(Level.FINE,
                    "This is the list of customers to notify (forward the received notification) : \n");
                }
                //----------------------------------------------------------------------------------------------------------
                
                EndpointReferenceType currentConsumerEdp = null;

                Boolean isValideSubsRef = true;

                if (subscriptionIds != null) {
                    for (final String subsId : subscriptionIds) {
                        isValideSubsRef = true;

                        // ------- Optimize notify forwards ---------
                        final EndpointReferenceType receivedSubsRef = msg.getSubscriptionReference();
                        if (receivedSubsRef != null) {
                            final ResourcesUuidType uuids = WsnSpecificTypeHelper.getResourcesUuidType(receivedSubsRef.getReferenceParameters());
                            if (uuids != null){

                                isValideSubsRef = false ;

                                final String subsRefUuid = uuids.getUuids().get(0);
                                this.logger.log(Level.FINE,"=== DEBUG ===: subsUuid received from Producer : " + subsRefUuid );

                                final List<MappingUuidThreesome> subsMappObjs = this.getThreesomeSubListBySubscribeResponseUuid(subsRefUuid);
                                //final List<RemoteSubscriptionProducerRefs> remoteSubsObjs = this.forwardedSubscriptionRefs.get(subsId);
                                if (/*remoteSubsObjs*/subsMappObjs != null) {
                                    for (final /*RemoteSubscriptionProducerRefs*/MappingUuidThreesome subsMappObjItem : subsMappObjs) {
                                        //final List<String> uuidsToCheck = WsnSpecificTypeHelper.getResourcesUuidType(remoteSubsObjItem.subscriptionReference.getReferenceParameters()).getUuids();
                                        isValideSubsRef = isValideSubsRef || (subsMappObjItem.getProducerSubscribeRespUuid().compareTo(subsRefUuid) == 0);
                                        // isValideSubsRef = this.forwardedSubscriptionRefs.get(subsId).contains(subsRefUuid);
                                        this.logger.log(Level.FINE,"SE-Notif DEBUG : subsUuid From Kpi : " + subsId );
                                    }
                                }
                            }
                        }


                        if (isValideSubsRef){
                            // ---------------------------------------
                           
                            currentConsumerEdp = this.wsnbSubscriptionManagerService
                            .getConsumerEdpRefOfSubscription(subsId);

                            // --- added by tdejean 13/10/2009 ------------
                            EndpointReferenceType currentSubscriptionRef = WSAddressingFactory.getInstance().newEndpointReferenceType();
                            currentSubscriptionRef.setAddress(currentConsumerEdp.getAddress());
                            
                            ReferenceParametersType ref = currentSubscriptionRef.newReferenceParameters();
                            currentSubscriptionRef.setReferenceParameters(ref);
                            
                            SOAParameterType soaParamOfSubs = WSNotificationExtensionFactory.getInstance().createSOAParameterType();
                            soaParam.setService(this.wsnbSubscriptionManagerService.getSubscriptionsManagerService());
                            soaParam.setInterface(this.wsnbSubscriptionManagerService.getSubscriptionsManagerInterface());
                            soaParam.setEndpoint(this.wsnbSubscriptionManagerService.getSubscriptionsManagerEdp());
                            
                            WsnSpecificTypeHelper.setSOAParameter(soaParamOfSubs, ref);
                            
                            final ResourcesUuidType rUuids = WSNotificationExtensionFactory
                                    .getInstance().createResourcesUuidType();
                            rUuids.addUuid(subsId);
                            WsnSpecificTypeHelper.setResourcesUuidType(rUuids, ref);

                            // msg.setSubscriptionReference(currentConsumerEdp);
                            msg.setSubscriptionReference(currentSubscriptionRef);
                            // -----end added 13/10/2009 --------------------

                            if (currentConsumerEdp != null) {
                                // Replace the Ws-Addressing address field with the
                                // subscribed consumer's one
                                newNotificationMsgOwner.setAddress(currentConsumerEdp.getAddress());

                                final SOAParameterType currentSoaParam = WsnSpecificTypeHelper.getSOAParameter(currentConsumerEdp.getReferenceParameters());
                                final QName targetService = currentSoaParam.getService();
                                this.logger.log(Level.FINE, "target consumer Service to notify : "
                                        + targetService + "\n");
                                final QName targetInterface = currentSoaParam.getInterface();
                                this.logger.log(Level.FINE, "target consumer interface to notify : "
                                        + targetService + "\n");
                                final String targetEndpoint = currentSoaParam.getEndpoint();
                                this.logger.log(Level.FINE, "target consumer Endpoint to notify : "
                                        + targetEndpoint);

                                // Replace Notification Sender Endpoint Reference
                                // with notification broker's one
                                msg.setProducerReference(newNotificationMsgOwner);
                                // ---- here, i forward received notifications to
                                // subscribed consumer
                                // Document requestAsDocument =
                                // WSNotificationFactory.getInstance().newWSNotificationWriter().writeNotify(request);
                                final Document requestAsDocument = WSNotificationWriter.getInstance(this.additonalnsAndPrefixMapping)
                                .writeNotify(request);
                                // this.forwardNotification(targetEdp,
                                // requestAsDocument);
                                this.forwardNotification(targetService, targetInterface,
                                        targetEndpoint, requestAsDocument);
                            }
                        }
                    }
                }
            }
        } catch (final WSNotificationException e) {
            this.logger.log(Level.WARNING, "[" + e.getClass().getSimpleName()
                    + "] -An error occurs during \"notify\" request proccessing");
            // e.printStackTrace();
        } catch (final WSAddressingException e) {
            this.logger.log(Level.WARNING, "[" + e.getClass().getSimpleName()
                    + "] -An error occurs during \"notify\" request proccessing");
            // e.printStackTrace();
        } catch (final WSNotificationExtensionException e) {
            this.logger.log(Level.WARNING, "[" + e.getClass().getSimpleName()
                    + "] -An error occurs during \"notify\" request proccessing");
            // e.printStackTrace();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.engine.test.BrokeredNotification#registerPublisher(com
     * .ebmwebsourcing.wsstar.notification.brokered.RegisterPublisher)
     */
    public RegisterPublisherResponse registerPublisher(final RegisterPublisher request)
    throws WSNotificationException, WSNotificationFault {

        RegisterPublisherResponse result = null;
        // ----- 1°/ Get existing SubscriptionRPs, from wsnbProducerService,
        // that match with the request's topicExpression
        // ----- 2°/ For each subscription, build a subscription request with
        // brokers endpoint ref as "consumerRef endpoint" of the request
        // ----- 3°/ send a new subscription to publisher, with the same filter
        // as )
        try{
            result = this.wsnbrRegistrationMgr.registerPublisher(request);

            // --- create a new Endpoint Reference to substitute to subscribe
            // msg 's one ---
            EndpointReferenceType newSubscribeMsgOwner;
            newSubscribeMsgOwner = WSAddressingFactory.getInstance()
            .newEndpointReferenceType();
            final ReferenceParametersType refParam = newSubscribeMsgOwner.newReferenceParameters();
            newSubscribeMsgOwner.setReferenceParameters(refParam);

            final SOAParameterType soaParam = WSNotificationExtensionFactory.getInstance().createSOAParameterType();
            soaParam.setEndpoint(this.wsnbProducerService.getNotificationProducerEdp());
            soaParam.setService(this.wsnbProducerService.getNotificationProducerService());
            soaParam.setInterface(this.wsnbProducerService.getNotificationProducerInterfaceQName());

            WsnSpecificTypeHelper.setSOAParameter(soaParam, refParam);

            newSubscribeMsgOwner.setAddress(SOAParameterType.DEFAULT_PETALS_ADDRESS);

            // ----------------------------------------------------------------------------------
            final EndpointReferenceType publisherProducerRef = request.getPublisherReference();

            final SOAParameterType pubSoaParam = WsnSpecificTypeHelper.getSOAParameter(publisherProducerRef.getReferenceParameters());

            final String targetEdp =pubSoaParam.getEndpoint();
            final QName targetInterface = pubSoaParam.getInterface();
            final QName targetService = pubSoaParam.getService();

            this.logger.log(Level.FINE,"target producer which a susbcribe request must be sent to (if subscriptions previously processed): ");
            this.logger.log(Level.FINE, "target producer service : " + targetService);
            this.logger.log(Level.FINE, "target producer interface : " + targetInterface);
            this.logger.log(Level.FINE, "target producer endpoint : " + targetEdp + "\n");

            final List<TopicExpressionType> topExprs = request.getTopic();
            for (final TopicExpressionType topicExprItem : topExprs) {
                
                final List<String> subsIds = this.wstopTopicsMgr.getSubscriptionIdsFromTopicsSet(
                        topicExprItem, false);
                if (subsIds != null && subsIds.size() > 0) {

                    this.logger.log(Level.FINE,"Some subscriptions have been found for topic expression : "
                            + topicExprItem.getContent());
                    for (final String subsIdItem : subsIds) {

                        // ------ Change uuids management 29/06/2009 ------
                        // ----- check if some subscribe request have already
                        // been forwarded
                        
                        //List<RemoteSubscriptionProducerRefs> fwdSubscriptionRefs = null;
                        //boolean isFirstsubscribeRequestForward = false;
                        //fwdSubscriptionRefs = this.forwardedSubscriptionRefs.get(subsIdItem);

                        //if (fwdSubscriptionRefs == null) {
                        //    fwdSubscriptionRefs = Collections.synchronizedList(new ArrayList<RemoteSubscriptionProducerRefs>());
                        //    isFirstsubscribeRequestForward = true;
                        //}
                        // ---------------------------------------------

                        final Subscribe subscriptionPayload = WSNotificationFactory.getInstance().createSubscribe();

                        final EndpointReferenceType subscribedConsumerEdp = this.wsnbSubscriptionManagerService.getConsumerEdpRefOfSubscription(subsIdItem);
                        newSubscribeMsgOwner.setAddress(subscribedConsumerEdp.getAddress());

                        subscriptionPayload.setConsumerReference(newSubscribeMsgOwner);

                        final FilterType filterToSet = this.wsnbSubscriptionManagerService.getFilterOfSubscription(subsIdItem);
                        subscriptionPayload.setFilter(filterToSet);

                        // ---- add susbcriptionPolicy from subscriptionResource
                        subscriptionPayload.setSubscriptionPolicy(this.wsnbSubscriptionManagerService.getPolicyOfSubscription(subsIdItem));

                        // ---- here, i forward received subscriptions to
                        // registered producer
                        final Document requestAsDocument = WSNotificationWriter.getInstance(this.additonalnsAndPrefixMapping).writeSubscribe(subscriptionPayload);

                        final SubscribeResponse currentSubsResp = this.forwardSubscriptionRequest(targetService, targetInterface, targetEdp, requestAsDocument);
                        if (currentSubsResp != null) {
                            final EndpointReferenceType returnedSubsRef = currentSubsResp.getSubscriptionReference();
 
                            /*fwdSubscriptionRefs.add(new RemoteSubscriptionProducerRefs(
                                    returnedSubsRef, publisherProducerRef,this.logger));

                            if (isFirstsubscribeRequestForward) {
                                this.forwardedSubscriptionRefs.put(subsIdItem, fwdSubscriptionRefs);
                            }*/
                                                            
                            // --- save and persist SubscribeResponses received from producer ---                            
                            String curForwSubsRespUuid = WsnSpecificTypeHelper.getResourcesUuidType(returnedSubsRef.getReferenceParameters()).getUuids().get(0);
                            this.receivedSubsResponses.put(curForwSubsRespUuid, currentSubsResp);   
                            this.persistProducerSubscribeResponse(currentSubsResp, curForwSubsRespUuid);
                            
                            // --- save and persist Subscription mapping threesome
                            MappingUuidThreesome newThreesome = 
                                new MappingUuidThreesome(subsIdItem,
                                    WsnSpecificTypeHelper.getResourcesUuidType(result.getPublisherRegistrationReference().getReferenceParameters()).getUuids().get(0),
                                    curForwSubsRespUuid);                            
                            this.subscriptionMappingList.add(newThreesome);                            
                            this.serializeSubscriptionMappingThreesome(newThreesome);
                            
                            // -- for debug --
                            this.displayUuidsMappingTable();
                            // --------------

                        }
                    }
                }
            }

        } catch (final WSAddressingException e) {
            this.logger.log(Level.WARNING, "[" + e.getClass().getSimpleName()
                    + "] -An error occurs during \"registerPublisher\" request proccessing");
            // e.printStackTrace();
        } catch (final WSNotificationExtensionException e) {
            this.logger.log(Level.WARNING, "[" + e.getClass().getSimpleName()
                    + "] -An error occurs during \"registerPublisher\" request proccessing");
            e.printStackTrace();
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @seecom.ebmwebsourcing.wsstar.notification.service.brokerednotification.
     * WsnbrPublisherRegistrationManager
     * #destroyRegistration(com.ebmwebsourcing.wsstar
     * .notification.brokered.DestroyRegistration)
     */
    public DestroyRegistrationResponse destroyRegistration(final DestroyRegistration request)
    throws WSNotificationException, WSNotificationFault {

        // --- Remove all references to the publisher in
        // forwardedSubscriptionRefs lists ...
        try {
            final List<String> uuidsToDestroy = WsnSpecificTypeHelper.getResourcesUuidType(request).getUuids();
            for (final String uuidItem : uuidsToDestroy) {
                    
                // remove "deprecated" - because registration removed - subscribe Response 
                List<MappingUuidThreesome> threesomesWithSubsRespToRemove = this.getThreesomeSubListByRegistrationUuid(uuidItem);
               
                for (MappingUuidThreesome threesomeItem : threesomesWithSubsRespToRemove) {
                    this.receivedSubsResponses.remove(threesomeItem.getProducerSubscribeRespUuid());                   
                    this.removePersistedSubscribeResponse(threesomeItem.getProducerSubscribeRespUuid());
                }
                
                // remove "deprecated" threesome 
                this.logger.log(Level.FINE,"- DestroyRegistration Request - remove existing threesomes containing " + uuidItem + " publisher registration uuid ...");                                                                           
                this.removeThreesomesByPublisherRegistrationUuid(uuidItem);              
                                
                // --- useful for debug 
                this.displayUuidsMappingTable();

                /*
                    final String prodRefToRemoveFromList = WsnSpecificTypeHelper.getSOAParameter(pubRef.getReferenceParameters()).getEndpoint();

                    //  final Collection<List<RemoteSubscriptionProducerRefs>> allSubscriptionList = this.forwardedSubscriptionRefs.values();
                    for (final Object element : allSubscriptionList) {
                        final List<RemoteSubscriptionProducerRefs> list = (List<RemoteSubscriptionProducerRefs>) element;

                        final List<RemoteSubscriptionProducerRefs> subsProdRefToRemove = Collections.synchronizedList(new ArrayList<RemoteSubscriptionProducerRefs>());
                        for (final RemoteSubscriptionProducerRefs remSubProRefItem : list) {
                            final EndpointReferenceType prodRef = remSubProRefItem.getProducerReference();
                            if (WsnSpecificTypeHelper.getSOAParameter(prodRef.getReferenceParameters()).getEndpoint().compareTo(prodRefToRemoveFromList) == 0) {
                                subsProdRefToRemove.add(remSubProRefItem);
                            }
                        }
                        list.removeAll(subsProdRefToRemove);
                    }  */


                //}
            }
            
        } catch (final WSNotificationExtensionException e) {
            //throw new WSNotificationException(e); 
            this.logger.log(Level.WARNING, "[" + e.getClass().getSimpleName()
                    + "] -An error occurs during \"DestroyRegistration\" request proccessing", e);
        }

        // -------------

        DestroyRegistrationResponse result;
        try {
            result = this.wsnbrRegistrationMgr.destroyRegistration(request);
        } catch (final WSNotificationExtensionException e) {
            throw new WSNotificationException(e);
        }
        // NOTE : The publisher that send the destroy request registration must
        // clear it's internal subscriptionManagerRP and subscriptionIds store
        // in its supported topicSet;

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.engine.test.BrokeredNotification#subscribe(com.ebmwebsourcing
     * .wsstar.notification.base.Subscribe)
     */
    public SubscribeResponse subscribe(final Subscribe request) throws WSNotificationException,WSNotificationFault {
        return this.subscribe(request, false, "");
    }
    
    
    private SubscribeResponse subscribe(final Subscribe request, boolean isRestart,String uuidSubscription) throws WSNotificationException,
    WSNotificationFault {
        SubscribeResponse result = null;
       
        String subsId;
        try {

            if (isRestart)            
                subsId = uuidSubscription;
            else {
                result = this.wsnbProducerService.subscribe(request);
                subsId = WsnSpecificTypeHelper.getResourcesUuidType(result.getSubscriptionReference().getReferenceParameters()).getUuids().get(0);
           
            }
                
            // ----- 1°/ Get existing PublisherRegistrationRPs, from
            // wsnbrRegistrationMgr, that match with the request's
            // topicExpression
            // ----- 2°/ For each registration substitute the request
            // "consumerRef endpoint" with the notification broker's one
            // ----- 3°/ Subscribe to each registered publisher previously found
            // by sending a subscribe request (use the result of 2°/)

            // --- create a new Endpoint Reference to substitute to subscribe
            // msg 's one ---
            EndpointReferenceType newSubscribeMsgOwner;
            newSubscribeMsgOwner = WSAddressingFactory.getInstance().newEndpointReferenceType();
            final ReferenceParametersType refParam = newSubscribeMsgOwner.newReferenceParameters();
            newSubscribeMsgOwner.setReferenceParameters(refParam);

            final SOAParameterType soaParam = WSNotificationExtensionFactory.getInstance().createSOAParameterType();

            soaParam.setEndpoint(this.wsnbProducerService.getNotificationProducerEdp());
            soaParam.setService(this.wsnbProducerService.getNotificationProducerService());
            soaParam.setInterface(this.wsnbProducerService.getNotificationProducerInterfaceQName());

            WsnSpecificTypeHelper.setSOAParameter(soaParam, refParam);

            // ----------------------------------------------------------------------------------

            // Get, from the "Topic manager", the subscriptionIds of the given
            // topic
            final List<String> registrationIds = this.wstopTopicsMgr.getRegistrationIdsFromTopicsSet(request.getFilter().getTopicExpression());

            // Get from the "Subscription manager" the associated consumers list
            this.logger.log(Level.FINE,"This is the list of Producer to susbcribe to (forward the received consumer subscription) : \n");
            EndpointReferenceType currentProducerEdp = null;

            //final List<RemoteSubscriptionProducerRefs> fwdSubscriptionRefs = Collections.synchronizedList(new ArrayList<RemoteSubscriptionProducerRefs>());
            
            if (registrationIds != null) {            
                for (final String regId : registrationIds) {

                    currentProducerEdp = this.wsnbrRegistrationMgr.getPublisherEdpRefOfRegistration(regId);                   
                    
                    if (currentProducerEdp != null) {
                        // Replace the Ws-Addressing address field with the
                        // subscribed consumer's one
                        newSubscribeMsgOwner.setAddress(currentProducerEdp.getAddress());

                        final SOAParameterType curSOAParam = WsnSpecificTypeHelper.getSOAParameter(currentProducerEdp.getReferenceParameters());

                        final QName targetService = curSOAParam.getService();
                        final QName targetInterface = curSOAParam.getInterface();
                        final String targetEdp = curSOAParam.getEndpoint();

                        this.logger.log(Level.FINE, "target producer Service to subscribe to : "  + targetService + "\n");
                        this.logger.log(Level.FINE, "target producer Interface to subscribe to : " + targetInterface + "\n");
                        this.logger.log(Level.FINE, "target producer Endpoint to subscribe to : " + targetEdp + "\n");

                        request.setConsumerReference(newSubscribeMsgOwner);

                        // ---- here, i forward received subscriptions to
                        // registred producer
                        final Document requestAsDocument = WSNotificationWriter.getInstance(this.additonalnsAndPrefixMapping)
                        .writeSubscribe(request);

                        final SubscribeResponse currentSubsResp = this.forwardSubscriptionRequest(
                                targetService, targetInterface, targetEdp, requestAsDocument);
                        
                        String curForwSubsRespUuid;
                        
                        if (currentSubsResp != null) {
                            final EndpointReferenceType returnedSubsRef = currentSubsResp.getSubscriptionReference();
                            //fwdSubscriptionRefs.add(new RemoteSubscriptionProducerRefs(
                            //        returnedSubsRef, currentProducerEdp, this.logger));
                            
                            // --- save SubscribeResponses received from producer ---                            
                            curForwSubsRespUuid = WsnSpecificTypeHelper.getResourcesUuidType(returnedSubsRef.getReferenceParameters()).getUuids().get(0);
                            this.receivedSubsResponses.put(curForwSubsRespUuid, currentSubsResp);
                            this.persistProducerSubscribeResponse(currentSubsResp, curForwSubsRespUuid);
                            // -------------------------------------------------------
                            
                            // ---- populate mapping list and persist ----
                            MappingUuidThreesome  newThreesome = new MappingUuidThreesome(subsId,regId,curForwSubsRespUuid);
                            this.subscriptionMappingList.add(newThreesome);
                            this.serializeSubscriptionMappingThreesome(newThreesome);
                            
                            // -- for debug --
                            this.displayUuidsMappingTable();
                        }
                    }                                     
                }                               
            }

            // --- for each subscription request processed, i must store the
            // endpointReference of
            // --- the forwarded subscription request response :
            //this.forwardedSubscriptionRefs.put(subsId, fwdSubscriptionRefs)            
            
            // -- for debug --
            //this.displaySubscriptionTable();
            // --------------

        } catch (final WSAddressingException e) {
            //throw new WSNotificationException(e); 
            this.logger.log(Level.WARNING, "[" + e.getClass().getSimpleName()
                    + "] -An error occurs during \"Subscribe\" request proccessing", e);
        } catch (final WSNotificationExtensionException e) {
            this.logger.log(Level.WARNING, "[" + e.getClass().getSimpleName()
                    + "] -An error occurs during \"Subscribe\" request proccessing", e);
            //throw new WSNotificationException(e);
        }
        // ---------------
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.engine.test.BrokeredNotification#unsubscribe(com.
     * ebmwebsourcing.wsstar.notification.base.Unsubscribe)
     */
    public UnsubscribeResponse unsubscribe(final Unsubscribe cancellation) throws WSNotificationException, WSNotificationFault {

        UnsubscribeResponse result = null;
        try {
            
            // -------- process request
            result = this.wsnbSubscriptionManagerService.unsubscribe(cancellation);            
            
            // ------- forward unsubscribe request to producers :            
            List<String> subscriptionIds = null;

            // ---- Get, from the subscriptionRP, the associated topicExpression
            // ---- Note that it must be done before process the unsubscribe
            // operation
            // ---- (As consequence of the unsubscribe operation, the
            // subscriptionRP
            // ---- is destroy)
            final ResourcesUuidType subsIdToRemove = WsnSpecificTypeHelper.getResourcesUuidType(cancellation);

            if (subsIdToRemove != null) {
                subscriptionIds = subsIdToRemove.getUuids();
            }
            
            // ----- 1°/ Get remaining subscription
            // ----- 2°/ If no subscription remains then get existing
            // PublisherRegistrationRPs from wsnbrRegistrationMgr that match
            // with the request's topicExpression
            // ----- and pause - or unsubscribe ? - any subscription previously
            // sent to the publisher

            // --- create a new Endpoint Reference to substitute to unsubscribe
            // msg 's one ---

            if (subscriptionIds != null) {
                for (final String subsIdItem : subscriptionIds) {
                                        
                    //final List<RemoteSubscriptionProducerRefs> remoteSubscriptionsRefs = this.forwardedSubscriptionRefs.get(subsId);
                    final List<MappingUuidThreesome> concernedThreesomes = this.getThreesomeSubListBySubscriptionUuid(subsIdItem);
                                                            
                    //if (remoteSubscriptionsRefs != null) {
                    if ((concernedThreesomes != null) && (concernedThreesomes.size()>0)) {
                        this.logger.log(Level.FINE,
                                "This is the list of Producer to unsusbcribe to (forward the received consumer unsubscribe requet) : \n");

                        EndpointReferenceType currentSubscriptionProducerEdp = null;
                        //for (final RemoteSubscriptionProducerRefs remSubProdRefItem : remoteSubscriptionsRefs) {
                        for (final MappingUuidThreesome threesomeItem : concernedThreesomes) {
                            
                            this.logger.log(Level.FINE,"---> [DEBUG] current Threesome used : " + threesomeItem.getConsumerSubcriptionUuid() + 
                                    " // " + threesomeItem.getPublisherRegistrationUuid() + " // " + threesomeItem.getProducerSubscribeRespUuid());
                            
                            //regRefItem = this.wsnbrRegistrationMgr.getPublisherEdpRefOfRegistration(threesomeItem.publisherRegistrationUuid);
                            currentSubscriptionProducerEdp = this.receivedSubsResponses.get(threesomeItem.getProducerSubscribeRespUuid()).getSubscriptionReference();
                            
                            final SOAParameterType regSoaParam = WsnSpecificTypeHelper.getSOAParameter(currentSubscriptionProducerEdp.getReferenceParameters());

                            final QName targetService = regSoaParam.getService();
                            final QName targetInterface =regSoaParam.getInterface();
                            final String targetEdp = regSoaParam.getEndpoint();
                            this.logger.log(Level.FINE,"target producer Service to Unsunsubscribe to : " + targetService + "\n");
                            this.logger.log(Level.FINE,"target producer Interface to Unsunsubscribe to : " + targetInterface + "\n");
                            this.logger.log(Level.FINE,"target producer Endpoint to Unsunsubscribe to : " + targetEdp + "\n");

                            // Replace Notification Sender Endpoint Reference
                            // with notification broker's one
                            // and forword the existaing registrationId
                            final Unsubscribe unsubscribeToForward = WSNotificationFactory.getInstance().createUnsubscribe();
                           
                            final ResourcesUuidType resourceUuids = WSNotificationExtensionFactory.getInstance().createResourcesUuidType();
                                                      
                            resourceUuids.addUuid(threesomeItem.getProducerSubscribeRespUuid());
                                                        
                            WsnSpecificTypeHelper.setResourcesUuidType(resourceUuids, unsubscribeToForward);

                            final com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType edpRef = WSAddressingFactory
                            .getInstance().newEndpointReferenceType();
                            //edpRef.setAddress(regRefItem.getAddress());
                            edpRef.setAddress(currentSubscriptionProducerEdp.getAddress());
                            
                            WsnSpecificTypeHelper.setEndpointReference(edpRef, unsubscribeToForward);

                            // ---- here, i forward received notifications to
                            // subscribed consumer
                            final Document requestAsDocument = WSNotificationWriter.getInstance(this.additonalnsAndPrefixMapping)
                            .writeUnsubscribe(unsubscribeToForward/* cancellation */);
                            // As the same registrationId is used the return is
                            // not necessary (?)
                            this.forwardUnsubscriptionRequest(targetService, targetInterface, targetEdp, requestAsDocument);
                            
                            // remove producer subscribe response from list
                            this.receivedSubsResponses.remove(threesomeItem.getProducerSubscribeRespUuid());
                            this.removePersistedSubscribeResponse(threesomeItem.getProducerSubscribeRespUuid());
                        }
                     // remove from uuids from forwardedSubscriptionRefs
                        this.removeThreesomesByConsumerSubscriptionUuid(subsIdItem);
                        // -- for debug --
                        this.displayUuidsMappingTable();
                    }                    
                }
            }

            // -- for debug --
            //this.displaySubscriptionTable();
            // --------------
            // ---------------
        } catch (final WSAddressingException e) {
            //throw new WSNotificationException(e);
            this.logger.log(Level.WARNING, "[" + e.getClass().getSimpleName()
                    + "] -An error occurs during \"unsubscribe\" request proccessing", e);
        } catch (final WSNotificationExtensionException e) {
            //throw new WSNotificationException(e); 
            this.logger.log(Level.WARNING, "[" + e.getClass().getSimpleName()
                    + "] -An error occurs during \"unsubscribe\" request proccessing", e);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @seecom.ebmwebsourcing.wsstar.notification.service.basenotification.
     * WsnbNotificationProducer
     * #getCurrentMessage(com.ebmwebsourcing.wsstar.notification
     * .base.GetCurrentMessage)
     */
    public GetCurrentMessageResponse getCurrentMessage(final GetCurrentMessage request)
    throws WSNotificationException, WSNotificationFault {
        return this.wsnbProducerService.getCurrentMessage(request);
    }

    /*
     * (non-Javadoc)
     * 
     * @seecom.ebmwebsourcing.wsstar.notification.service.basenotification.
     * WsnbSubscriptionManager
     * #pauseSubscription(com.ebmwebsourcing.wsstar.notification
     * .base.PauseSubscription)
     */
    /*
     * public PauseSubscriptionResponse pauseSubscription(PauseSubscription
     * request) throws WSNotificationException, WSNotificationFault { return
     * this.wsnbSubscriptionManagerService.pauseSubscription(request); }
     */
    /*
     * (non-Javadoc)
     * 
     * @seecom.ebmwebsourcing.wsstar.notification.service.basenotification.
     * WsnbSubscriptionManager
     * #renew(com.ebmwebsourcing.wsstar.notification.base.Renew)
     */
    public RenewResponse renew(final Renew request) throws WSNotificationException, WSNotificationFault {
        RenewResponse result = null;        
        try {
            result = this.wsnbSubscriptionManagerService.renew(request);
        } catch (WSNotificationExtensionException e) {
            throw new WSNotificationException(e); 
        }        
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @seecom.ebmwebsourcing.wsstar.notification.service.basenotification.
     * WsnbSubscriptionManager
     * #resumeSubscription(com.ebmwebsourcing.wsstar.notification
     * .base.ResumeSubscription)
     */
    /*
     * public ResumeSubscriptionResponse resumeSubscription( ResumeSubscription
     * request) throws WSNotificationException, WSNotificationFault { return
     * this.wsnbProducerService.resumeSubscription(request); }
     */
    /*
     * (non-Javadoc)
     * 
     * @seecom.ebmwebsourcing.wsstar.notification.service.basenotification.
     * WsnbCreatePullPoint
     * #createPullPoint(com.ebmwebsourcing.wsstar.notification
     * .base.CreatePullPoint)
     */
    public CreatePullPointResponse createPullPoint(final CreatePullPoint request)
    throws WSNotificationException, WSNotificationFault {
        return this.wsnbCreatePullPointService.createPullPoint(request);
    }

    // #################################################################################
    // ################# Others technical methods (presistence,  ...)###################
    
    /**
     *  used for debug : display current mapping list
     */
    private void displayUuidsMappingTable(){
        this.logger.log(Level.FINE,"----- This is the \"Uuid Threesomes Table\" -----\n" +
                " schema : \"UuidOfConsumerSubscription\" // \"UuidOfPublisherRegistration\" // \"UuidOfProducerSubscribeResp\"");
        
        for (final MappingUuidThreesome mappingItem : this.subscriptionMappingList) {
        
            this.logger.log(Level.FINE,"---> Threesome : " + mappingItem.getConsumerSubcriptionUuid() + 
                    " // " + mappingItem.getPublisherRegistrationUuid() + " // " + mappingItem.getProducerSubscribeRespUuid());
          
            this.logger.log(Level.FINE," --- End of List of remote subscription Uuids (returned by Producers) ---\n ");
        
        }
        this.logger.log(Level.FINE,"----- --------------End Of  \"Uuid Threesomes Table\" display --------------- -----\n\n");
    }
    
    // ############# Forwarded Subscribe Responses persistence util : ###########
          
    /**
     *     
     * @param producerSubscibeRespUuid
     * @return
     */
    public List<MappingUuidThreesome> getThreesomeSubListBySubscriptionUuid(String consSubscriptionUuid){

        List<MappingUuidThreesome> result = new ArrayList<MappingUuidThreesome>();

        String currentConsSubscriptionUuid ; 
        for (MappingUuidThreesome subscriptionMappingItem : this.subscriptionMappingList) {           
            currentConsSubscriptionUuid = subscriptionMappingItem.getConsumerSubcriptionUuid();            
            if (currentConsSubscriptionUuid.compareTo(consSubscriptionUuid) == 0)
                result.add(subscriptionMappingItem);
        }        
        return result;                    
    }

    /**
     * 
     * @param consSubscriptionUuid
     * @return
     */
    public List<MappingUuidThreesome> getThreesomeSubListByRegistrationUuid(String producerRegistrationUuid){

        List<MappingUuidThreesome> result = new ArrayList<MappingUuidThreesome>();

        String currentProdRegistrationUuid ;         
        for (MappingUuidThreesome subscriptionMappingItem : this.subscriptionMappingList) {           
            currentProdRegistrationUuid = subscriptionMappingItem.getPublisherRegistrationUuid();            
            if (currentProdRegistrationUuid.compareTo(producerRegistrationUuid) == 0)
                result.add(subscriptionMappingItem);
        }        
        return result;                    
    }
    
    /**
     * 
     * @param prodSubscribeRespUuid
     * @return
     */
    public List<MappingUuidThreesome> getThreesomeSubListBySubscribeResponseUuid(String prodSubscribeRespUuid){

        List<MappingUuidThreesome> result = new ArrayList<MappingUuidThreesome>();

        String currentProdSubsRespUuid ; 
        for (MappingUuidThreesome subscriptionMappingItem : this.subscriptionMappingList) {           
            currentProdSubsRespUuid = subscriptionMappingItem.getProducerSubscribeRespUuid();            
            if (currentProdSubsRespUuid.compareTo(prodSubscribeRespUuid) == 0)
                result.add(subscriptionMappingItem);
        }        
        return result;                    
    }
    
    /**
     * 
     * @param consumerSubscriptionUuid
     * @throws WSNotificationException 
     */
    public void removeThreesomesByConsumerSubscriptionUuid(String consumerSubscriptionUuid) throws WSNotificationException{
        List<MappingUuidThreesome> threesomesToRemove = new ArrayList<MappingUuidThreesome>();
        
        for (MappingUuidThreesome mappingUuidThreesomeItem : this.subscriptionMappingList) {
            if (mappingUuidThreesomeItem.getConsumerSubcriptionUuid().compareTo(consumerSubscriptionUuid) == 0){
             
                this.logger.log(Level.FINE,"---> Threesome deleted : " + mappingUuidThreesomeItem.getConsumerSubcriptionUuid() + 
                        " // " + mappingUuidThreesomeItem.getPublisherRegistrationUuid() + " // " + mappingUuidThreesomeItem.getProducerSubscribeRespUuid());
               
                threesomesToRemove.add(mappingUuidThreesomeItem);                
            }
        }
        
        this.subscriptionMappingList.removeAll(threesomesToRemove);  
        // remove serialized threesomes
        this.removeSerializedSubscriptionMappingThreesome(threesomesToRemove);         
    }
    
    /**
     * 
     * @param publisherResgistrationUuid
     * @throws WSNotificationException 
     */
    public void removeThreesomesByPublisherRegistrationUuid(String publisherResgistrationUuid) throws WSNotificationException{
        List<MappingUuidThreesome> threesomesToRemove = new ArrayList<MappingUuidThreesome>();

        for (MappingUuidThreesome mappingUuidThreesomeItem : this.subscriptionMappingList) {
            if (mappingUuidThreesomeItem.getPublisherRegistrationUuid().compareTo(publisherResgistrationUuid) == 0){
        
                this.logger.log(Level.FINE,"---> Threesome deleted : " + mappingUuidThreesomeItem.getConsumerSubcriptionUuid() + 
                        " // " + mappingUuidThreesomeItem.getPublisherRegistrationUuid() + " // " + mappingUuidThreesomeItem.getProducerSubscribeRespUuid());                
                threesomesToRemove.add(mappingUuidThreesomeItem);
            }
        }
        
        this.subscriptionMappingList.removeAll(threesomesToRemove);
         
        // remove serialized threesomes
        this.removeSerializedSubscriptionMappingThreesome(threesomesToRemove);
    }
        
    /**
     * 
     * @param publisherResgistrationUuid
     * @throws WSNotificationException 
     */
    public void removeThreesomesByProducerSubscribeResponseUuid(String producerSubscribeResponseUuid) throws WSNotificationException{
        List<MappingUuidThreesome> threesomesToRemove = new ArrayList<MappingUuidThreesome>();

        for (MappingUuidThreesome mappingUuidThreesomeItem : this.subscriptionMappingList) {
            if (mappingUuidThreesomeItem.getProducerSubscribeRespUuid().compareTo(producerSubscribeResponseUuid) == 0){
        
                this.logger.log(Level.FINE,"---> Threesome deleted : " + mappingUuidThreesomeItem.getConsumerSubcriptionUuid() + 
                        " // " + mappingUuidThreesomeItem.getPublisherRegistrationUuid() + " // " + mappingUuidThreesomeItem.getProducerSubscribeRespUuid());                
                threesomesToRemove.add(mappingUuidThreesomeItem);
            }
        }
        
        this.subscriptionMappingList.removeAll(threesomesToRemove);
         
        // remove serialized threesomes
        this.removeSerializedSubscriptionMappingThreesome(threesomesToRemove);
    }
    
    // ----------------- "SubscriptionMappingList" serialize/deserialize methods -------------
    
    public void initSubscriptionMappingThreesomeFolder(){
        
        if (this.subscriptionMappingFolder == null) {
            throw new NullPointerException("Persistence folder for \"Subscription Mapping Threesome\" must not be null");
        }
        
        if (!this.subscriptionMappingFolder.exists()) {
            this.subscriptionMappingFolder.mkdirs();
        }                
    }    
    
    private void serializeSubscriptionMappingThreesome(MappingUuidThreesome threesome) throws WSNotificationException{

        String associatedFileName = threesome.getConsumerSubcriptionUuid() +
        threesome.getProducerSubscribeRespUuid() + 
        threesome.getPublisherRegistrationUuid() + ".serial";
        try{
            FileOutputStream fos = new FileOutputStream(this.subscriptionMappingFolder.getAbsolutePath() +
                    File.separatorChar + associatedFileName);
            ObjectOutputStream oos= new ObjectOutputStream(fos);
            try {           
                oos.writeObject(threesome);            
                oos.flush();
                this.logger.log(Level.FINE,threesome + " have been serialized ");
            } finally {               
                try {
                    oos.close();
                } finally {
                    fos.close();
                }
            }            
        } catch(IOException ioe) {
            throw new WSNotificationException("Serialization of threesome \"" + associatedFileName + "\" failed !",ioe);
            //ioe.printStackTrace();
        }
    }

    private void removeSerializedSubscriptionMappingThreesome(List<MappingUuidThreesome> threesomes) throws WSNotificationException{

        for (MappingUuidThreesome threesomeItem : threesomes) {

            String associatedFileName = threesomeItem.getConsumerSubcriptionUuid() +
            threesomeItem.getProducerSubscribeRespUuid() + 
            threesomeItem.getPublisherRegistrationUuid() + ".serial";

            this.logger.log(Level.FINE,"remove mapping threesome \"" + associatedFileName +"\" ...");

            File subscriptionToRemove = new File(this.subscriptionMappingFolder.getAbsoluteFile(),associatedFileName);
            if (subscriptionToRemove.exists())
                if (!subscriptionToRemove.delete())
                    throw new WSNotificationException("Deletion of threesome \"" + associatedFileName + "\" failed !");
        }  
    }
    
    private List<MappingUuidThreesome> deserializeSubscriptionMappingThreesomes() throws WSNotificationException{
      
        this.logger.log(Level.FINE,"Deserialize threesomes ...");
        
        File[] serializedThreesomesAsFile = this.subscriptionMappingFolder.getAbsoluteFile().listFiles();
        
        List<MappingUuidThreesome> result = new ArrayList<MappingUuidThreesome>();
        try {
            
            File currentFile = null;
            for (int i=0; i<serializedThreesomesAsFile.length;i++ /*String serializedItem : serializedThreesomes*/) {
               
                currentFile = serializedThreesomesAsFile[i].getAbsoluteFile();
                this.logger.log(Level.FINE,"current serialized threesomes :" + currentFile);
                
                if (!currentFile.exists())
                    this.logger.log(Level.FINE," /!\\ Warning File does not exist !! !!!");
                else{
                    FileInputStream fis = new FileInputStream(currentFile);
                    ObjectInputStream ois= new ObjectInputStream(fis);
                    try {   
                        result.add((MappingUuidThreesome) ois.readObject()); 
                    } finally {
                        try {
                            ois.close();
                        } finally {
                            fis.close();
                        }
                    }
                }
            }
        } catch(IOException ioe) {
            throw new WSNotificationException("Deserialization of threesomes failed !",ioe);
            //ioe.printStackTrace();
        } catch(ClassNotFoundException cnfe) { 
            throw new WSNotificationException("Deserialization of threesomes failed !",cnfe);
            //cnfe.printStackTrace();
        }
        this.logger.log(Level.FINE,"Threesomes deserialized !");
        return result;        
    }
    
    // ############################################################################################
    // ~~~~~~~~~~~~~~~~~ in/out persistence methods for "SubscribeResponse" obj ~~~~~~~~~~~~~~~~~~~    
    
    public void initProducerSubscribeResponseFolder(){
        
        if (this.forwardedSubscribeResponsesFolder == null) {
            throw new NullPointerException("Persistence folder for \"Producer Subscribe Responses\" must not be null");
        }
        
        if (!this.forwardedSubscribeResponsesFolder.exists()) {
            this.forwardedSubscribeResponsesFolder.mkdirs();
        }
                
    }
    
    /**
     * 
     * @param subscription
     * @throws Exception
     */
    private void persistProducerSubscribeResponse(SubscribeResponse subscribeResp, String uuid) throws WSNotificationException {
        
        this.logger.log(Level.FINE,"persiste subscribe Response \"" + uuid +"\" ...");
        
        File persist = new File(this.forwardedSubscribeResponsesFolder.getAbsoluteFile(),uuid);
        if (persist == null) {
            throw new WSNotificationException("The persistance file can not be created");
        }
        if (!persist.exists()) {
            try {
                persist.createNewFile();
            } catch (IOException e) {
                throw new WSNotificationException(e);
            }
        }        
        WSNotificationWriter.getInstance().writeSubscribeResponse(subscribeResp, persist);
    }
    
    /**
     * 
     * @return
     * @throws WSNotificationException 
     */
    private Map<String,SubscribeResponse> getSubscribeResponseToRestore() throws WSNotificationException{
       
        this.logger.log(Level.FINE,"restore subscribe Responses ...");
        
        Map<String, SubscribeResponse> result = new HashMap<String, SubscribeResponse>();
        if (this.forwardedSubscribeResponsesFolder.exists()){
            File[] persistedSubsriptionFiles = this.forwardedSubscribeResponsesFolder.getAbsoluteFile().listFiles();
            for (int i = 0; i < persistedSubsriptionFiles.length; i++) {
                result.put(persistedSubsriptionFiles[i].getName(),
                        WSNotificationReader.getInstance().readSubscribeResponse(persistedSubsriptionFiles[i]));
            }
        }           
        return result;
    }
    
    /**
     * 
     * @param uuid
     * @throws WSNotificationException
     */
    private void removePersistedSubscribeResponse(String uuid) throws WSNotificationException{  
        
        this.logger.log(Level.FINE,"remove subscribe Response \"" + uuid +"\" ...");
        
        File subscriptionToRemove = new File(this.forwardedSubscribeResponsesFolder.getAbsoluteFile(),uuid);
        if (subscriptionToRemove.exists())
            if (!subscriptionToRemove.delete())
                throw new WSNotificationException("Deletion of subscription \"" + uuid + "\" failed !");
    }
    // #########################################################################
    
}
