/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals;

import javax.jbi.JBIException;

import org.ow2.petals.component.framework.se.AbstractServiceEngine;

import com.ebmwebsourcing.wsstar.notification.definition.inout.WSNotificationReader;
import com.ebmwebsourcing.wsstar.notification.definition.inout.WSNotificationWriter;
import com.ebmwebsourcing.wsstar.notification.definition.utils.WSNotificationException;
import com.ebmwebsourcing.wsstar.notification.service.fault.WSNotificationFault;

/**
 * The petals-se-notification Service Engine.
 * 
 * <br>
 * <b>NOTE : </b>This class has to be used only if the component developer wants
 * to customize the main component class. In general, using the
 * org.objectweb.petals.component.framework.se.DefaultServiceEngine class is
 * enough. If so, change the value in the JBI descriptor file.
 * 
 * @author Gardes Frederic
 * @author tdejean - eBM Websourcing -
 * 
 */
public class ServiceEngine extends AbstractServiceEngine {

    private BrokeredNotificationService wsnbrBrokerService;

    private WSNotificationReader wsnReader = null;

    private WSNotificationWriter wsnWriter = null;

    public BrokeredNotificationService getService() {
        return this.wsnbrBrokerService;
    }

    public WSNotificationReader getWsnReader() {
        return this.wsnReader;
    }

    public WSNotificationWriter getWsnWriter() {
        return this.wsnWriter;
    }

    @Override
    protected void doInit() throws JBIException {
        super.doInit();
        this.getLogger().info("init WS-Notification service");

        final String[] wsdlNsPrefPairs = this.getNativeWsdl().getDescription().getNamespaces()
        .getPreDeclaredNamespaceUris2();

        
        this.wsnReader = WSNotificationReader.getInstance();
        this.wsnWriter = WSNotificationWriter.getInstance(wsdlNsPrefPairs);
        
        this.wsnbrBrokerService = new BrokeredNotificationService(this.getLogger(), this,
                wsdlNsPrefPairs);

        this.wsnbrBrokerService.initProducerSubscribeResponseFolder();
        this.wsnbrBrokerService.initSubscriptionMappingThreesomeFolder();
    }
    
    @Override
    protected void doStart() throws JBIException {       
        super.doStart();
        try {
            this.wsnbrBrokerService.restorePreviousState();         
            
        } catch (WSNotificationException e) {
              e.printStackTrace();
        } catch (WSNotificationFault e) {
            e.printStackTrace();
        }
    }
    
    @Override
    protected void doStop() throws JBIException {
        // TODO Auto-generated method stub
        super.doStop();        
        
        try {
            this.wsnbrBrokerService.unsubscribeToAllProducers();
        } catch (WSNotificationException e) {
            e.printStackTrace();
        }
    }
    
    @Override
    protected void doShutdown() throws JBIException {
        this.wsnbrBrokerService = null;
        this.wsnReader = null;
        this.wsnWriter = null;

        super.doShutdown();
    }
}
